/*
 *************************************************************************
 *
 * serdisp_specific_l4m.c
 * routines for controlling displays controlled by linux4media usb-interface
 *
 * supported:
 * - linux4media E-5i-USB interface with RC-LCD-B 128x64 LCD
 *
 *************************************************************************
 *
 * copyright (C) 2007  wolfgang astleitner
 * email     mrwastl@users.sourceforge.net
 *
 *************************************************************************
 * This program is free software; you can redistribute it and/or modify   
 * it under the terms of the GNU General Public License as published by   
 * the Free Software Foundation; either version 2 of the License, or (at  
 * your option) any later version.                                        
 *                                                                        
 * This program is distributed in the hope that it will be useful, but    
 * WITHOUT ANY WARRANTY; without even the implied warranty of             
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU      
 * General Public License for more details.                               
 *                                                                        
 * You should have received a copy of the GNU General Public License      
 * along with this program; if not, write to the Free Software            
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA              
 * 02111-1307, USA.  Or, point your browser to                            
 * http://www.gnu.org/copyleft/gpl.html                                   
 *************************************************************************
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>

#include "serdisplib/serdisp_connect.h"
#include "serdisplib/serdisp_tools.h"
#include "serdisplib/serdisp_messages.h"


/* #define OPT_USEOLDUPDATEALGO */

/* different display types/models supported by this driver */
#define DISPID_L4ME5I       1


serdisp_options_t serdisp_l4m_options[] = {
   /*  name          aliasnames   min    max mod int defines  */
   {  "INVERT",      "",            0,    -1,  1, 1,  ""}
  ,{  "RAWCMD",      "",            0,   255,  1, 1,  ""}      /* for development - internal use only */
};


/* internal typedefs and functions */

static void serdisp_l4m_init        (serdisp_t*);
static void serdisp_l4m_update      (serdisp_t*);
static int  serdisp_l4m_setoption   (serdisp_t*, const char*, long);
static void serdisp_l4m_close       (serdisp_t*);


/* main functions */


/* *********************************
   serdisp_t* serdisp_l4m_setup(sdcd, dispname, optionstring)
   *********************************
   sets up a display descriptor fitting to dispname and extra
   *********************************
   sdcd             ... output device handle (not used in here)
   dispname         ... display name (case-insensitive)
   optionstring     ... option string containing individual options
   *********************************
   returns a display descriptor
*/
serdisp_t* serdisp_l4m_setup(const serdisp_CONN_t* sdcd, const char* dispname, const char* optionstring) {
  serdisp_t* dd;

  if (! (dd = (serdisp_t*)sdtools_malloc(sizeof(serdisp_t)) ) ) {
    sd_error(SERDISP_EMALLOC, "%s(): cannot allocate display descriptor", __func__);
    return (serdisp_t*)0;
  }
  memset(dd, 0, sizeof(serdisp_t));

  /* assign dd->dsp_id */
  if (serdisp_comparedispnames("L4ME5I", dispname))
    dd->dsp_id = DISPID_L4ME5I;
  else {  /* should not occur */
    sd_error(SERDISP_ENOTSUP, "display '%s' not supported by serdisp_l4m_ssdoled.c", dispname);
    return (serdisp_t*)0;
  }

  /* default setup for function pointers */
  dd->fp_init           = &serdisp_l4m_init;
  dd->fp_update         = &serdisp_l4m_update;
  dd->fp_close          = &serdisp_l4m_close;
  dd->fp_setoption      = &serdisp_l4m_setoption;
  dd->fp_setpixel       = &sdtools_generic_setpixel;
  dd->fp_getpixel       = &sdtools_generic_getpixel;


  /* per display settings */

  dd->width             = 128;
  dd->height            = 64;
  dd->depth             = 1;
  dd->feature_contrast  = 0;
  dd->feature_backlight = 0;
  dd->feature_invert    = 1;

  dd->dsparea_width     = 58840;     /* active viewing area in micrometres              */
  dd->dsparea_height    = 35150;     /*   according to datasheet of Powertip PG12864WRF */

  /* max. delta for optimised update algorithm */
  dd->optalgo_maxdelta  = 56;

  /* finally set some non display specific defaults */

  dd->curr_rotate       = 0;         /* unrotated display */
  dd->curr_invert       = 0;         /* display not inverted */

  /*
  dd->min_contrast      = 0x1;
  dd->max_contrast      = 0xA;
  */

  /* supported output devices */
  dd->connection_types  = SERDISPCONNTYPE_PARPORT | SERDISPCONNTYPE_HIDDEV;

  serdisp_setupstructinfos(dd, 0, 0, serdisp_l4m_options);

  /* parse and set options */
  if (serdisp_setupoptions(dd, dispname, optionstring) ) {
    free(dd->specific_data);
    free(dd);
    dd = 0;
    return (serdisp_t*)0;
  }

  return dd;
}



/* *********************************
   void serdisp_l4m_init(dd)
   *********************************
   initialise a linux4media usb-based display
   *********************************
   dd     ... display descriptor
*/
void serdisp_l4m_init(serdisp_t* dd) {
  SDCONN_write(dd->sdcd, 0x10000 | 0x01, 0);  /* init lcd */
  SDCONN_write(dd->sdcd, 0x10000 | 0x02, 0);  /* clear lcd */
  SDCONN_write(dd->sdcd, 0x10000 | 0x05, 0);  /* set normal display */
  SDCONN_write(dd->sdcd, 0x10000 | 0x06, 0);  /* autoadjust polarity (== bind pol. to empty screen) */
  sd_debug(2, "%s(): done with initialising", __func__);
}



/* *********************************
   void serdisp_l4m_update(dd)
   *********************************
   updates the display using display-buffer scrbuf+scrbuf_chg
   *********************************
   dd     ... display descriptor
   *********************************

   the display is redrawn using a time-saving algorithm:
*/
void serdisp_l4m_update(serdisp_t* dd) {
  int x;
  int page;
  int pages = (dd->height+7)/8;


#ifdef OPT_USEOLDUPDATEALGO
  for (page = 0; page < pages; page++) {
    for(x = 0; x < dd->width; x++) {
      if ((x % 56) == 0) {
        SDCONN_commit(dd->sdcd);
        SDCONN_write(dd->sdcd, 0x12, 0);
        SDCONN_write(dd->sdcd, page, 0);
        SDCONN_write(dd->sdcd, x,    0);
        SDCONN_write(dd->sdcd, (x >= 2*56) ? 16 : 56, 0);
      }
      SDCONN_write(dd->sdcd, dd->scrbuf[ dd->width * page  +  x], 0);
    }
  }
  SDCONN_commit(dd->sdcd);
  
#else /* OPT_USEOLDUPDATEALGO */

  /* display is drawn using an optimising algorithm which tries to only send as few data as possible to the display */

  int col, col_start;
  byte data;
  
  col_start = 0;  /* start column (if previous page has already set some columns in current page) */

  for (page = 0; page < pages; page++) {

    col = col_start;

    while (col < dd->width) {
      if ( dd->scrbuf_chg[ col + dd->width *(page/8)] & ( 1 << (page%8)) ) {
        int endm = dd->optalgo_maxdelta;

        if (page == 7 && col > (dd->width - dd->optalgo_maxdelta) )
          endm = dd->width - col;

        SDCONN_commit(dd->sdcd);
        SDCONN_write(dd->sdcd, 0x12, 0);
        SDCONN_write(dd->sdcd, page, 0);
        SDCONN_write(dd->sdcd, col, 0);
        SDCONN_write(dd->sdcd, endm, 0);

        for (x = col ; x < col + endm; x++) {
          int x_i = x;
          int page_i = page;

          if (x >= dd->width) {
            x_i = x - dd->width;
            page_i = page + 1;
          }

          data = dd->scrbuf[x_i + dd->width * page_i ];

          if (dd->curr_invert && !(dd->feature_invert))
            data = ~data;

          SDCONN_write(dd->sdcd, data, 0);
          dd->scrbuf_chg[ x_i + dd->width * (page_i/8)] &= 0xFF - (1 << (page_i%8)) ;
        }
        SDCONN_commit(dd->sdcd);

        col += endm;

        col_start = (col >= dd->width) ? (col - dd->width) : 0;
      } else {
        col++;
      }

    }
  }
#endif /* OPT_USEOLDUPDATEALGO */

  /* add an extra NOP to avoid erraneous pixels when releasing parport */
  SDCONN_commit(dd->sdcd); /* if streaming: be sure that every data is transmitted */
}





/* *********************************
   int serdisp_l4m_setoption(dd, option, value)
   *********************************
   change a display option
   *********************************
   dd      ... display descriptor
   option  ... name of option to change
   value   ... value for option
*/
int serdisp_l4m_setoption(serdisp_t* dd, const char* option, long value) {

  if (dd->feature_invert && serdisp_compareoptionnames(dd, option, "INVERT") ) {
    if (value < 2) 
      dd->curr_invert = (int)value;
    else
      dd->curr_invert = (dd->curr_invert) ? 0 : 1;
    /* 0x04: invers mode;  0x05: normal mode */
    SDCONN_write(dd->sdcd, 0x10000 | ((dd->curr_invert) ? 0x04 : 0x05), 0);
#if 0   /* changing 0x09 (adjustment / contrast) is no good idea */
  } else if (dd->feature_contrast && serdisp_compareoptionnames(dd, option, "CONTRAST" )) {
    if (value < 0 || value > MAX_CONTRASTSTEP)
      value = MAX_CONTRASTSTEP;

    dd->curr_contrast = (((dd->max_contrast - dd->min_contrast) / MAX_CONTRASTSTEP) * (int)value) + dd->min_contrast;

    SDCONN_write(dd->sdcd, 0x09, 0);
    SDCONN_write(dd->sdcd, dd->curr_contrast, 0);
    SDCONN_commit(dd->sdcd);
#endif
  } else if (serdisp_compareoptionnames(dd, option, "RAWCMD")) {
    fprintf(stderr, "val: 0x%02x\n", (byte)(0xFF & value));
    SDCONN_write(dd->sdcd, 0x10000 | ((byte)(0xFF & value)), 0);
  } else {
    /* option not found here: try generic one in calling serdisp_setoption(); */
    return 0;
  }
  return 1;
}



/* *********************************
   void serdisp_l4m_close(dd)
   *********************************
   close (switch off) display
   *********************************
   dd     ... display descriptor
*/
void serdisp_l4m_close(serdisp_t* dd) {
  serdisp_clear(dd);
  SDCONN_write(dd->sdcd, 0x10000 | 0x05, 0); /* force normal mode */
}





